/*
* Copyright (c) 2022 – 2023 Texas Instruments Incorporated
*
* All rights reserved not granted herein.
* Limited License. 
*
* Texas Instruments Incorporated grants a world-wide, royalty-free,
* non-exclusive license under copyrights and patents it now or hereafter
* owns or controls to make, have made, use, import, offer to sell and sell ("Utilize")
* this software subject to the terms herein.  With respect to the foregoing patent
* license, such license is granted  solely to the extent that any such patent is necessary
* to Utilize the software alone.  The patent license shall not apply to any combinations which
* include this software, other than combinations with devices manufactured by or for TI (“TI Devices”). 
* No hardware patent is licensed hereunder.
*
* Redistributions must preserve existing copyright notices and reproduce this license (including the
* above copyright notice and the disclaimer and (if applicable) source code license limitations below)
* in the documentation and/or other materials provided with the distribution
*
* Redistribution and use in binary form, without modification, are permitted provided that the following
* conditions are met:
*
*   * No reverse engineering, decompilation, or disassembly of this software is permitted with respect to any
*     software provided in binary form.
*   * any redistribution and use are licensed by TI for use only with TI Devices.
*   * Nothing shall obligate TI to provide you with source code for the software licensed and provided to you in object code.
*
* If software source code is provided to you, modification and redistribution of the source code are permitted
* provided that the following conditions are met:
*
*   * any redistribution and use of the source code, including any resulting derivative works, are licensed by
*     TI for use only with TI Devices.
*   * any redistribution and use of any object code compiled from the source code and any resulting derivative
*     works, are licensed by TI for use only with TI Devices.
*
* Neither the name of Texas Instruments Incorporated nor the names of its suppliers may be used to endorse or
* promote products derived from this software without specific prior written permission.
*
* DISCLAIMER.
*
* THIS SOFTWARE IS PROVIDED BY TI AND TI’S LICENSORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING,
* BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
* IN NO EVENT SHALL TI AND TI’S LICENSORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
* OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
* OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*/

module spi_master(

	input logic spi_clk,		//free running spi FSM clock		
	input logic SPI_EN,			//make it logic HIGH to enable the FSM		

	input logic tx_trn, 		//TX TRIGGER: 0 to 1 transition indicates write SPI data to DUT			 
	input logic rx_trn, 		//RX TRIGGER: 0 to 1 transition indicates read SPI data from DUT		 
	input logic [7:0] addr,		//DUT's register address		
	input logic [15:0] wr_data,	//DUT's register write data			

	input logic SPI_MISO,		//DUT's SDO		
	input logic rst,			//active high reset		

	output logic SPI_CS_Z,			//DUT's CSz		
	output logic SPI_MOSI,			//DUT's SDI	
	output logic SPI_SCLK,			//DUT's SCLK		
	output logic SPI_BUSY,			//FSM busy indicator
	output logic SPI_READ_DONE, 	//FSM read done indicator
	output logic [23:0] read_data 	//Deserialzed SPI_MISO

);

logic [4:0] tx_count;
logic counter_en;
logic read_mode_en;
logic spi_clk_out;
logic cs_z;
logic [23:0] tx_buff;
logic [23:0] tx_buff1;
logic [23:0] rx_buff;
logic tx_out;
logic read_done_flag;
logic trn;
logic fsm_start;
logic [2:0] fsm_start_count;
logic [23:0] temp;
logic busy;
logic read_done;
logic st_trn;
logic tx_trn_pulse;
logic rx_trn_pulse;
logic tx_trn_delay;
logic rx_trn_delay;

// tx_pulse generator
always @(posedge spi_clk or posedge rst) begin 
	if (rst) begin
		tx_trn_delay <= 1'b0;
	end 
	else begin
		tx_trn_delay <= tx_trn;
	end 
end 

assign tx_trn_pulse = tx_trn & (~tx_trn_delay);

// rx_pulse generator
always @(posedge spi_clk or posedge rst) begin 
	if (rst) begin
		rx_trn_delay <= 1'b0;
	end 
	else begin
		rx_trn_delay <= rx_trn;
	end 
end 

assign rx_trn_pulse = rx_trn & (~rx_trn_delay);
assign trn = (tx_trn_pulse | rx_trn_pulse) & (!busy);
assign tx_buff = {addr,wr_data};
 
always @(posedge spi_clk or posedge rst or posedge trn) begin 
	if (rst) begin 
		fsm_start_count <=  3'd0;
	end 
	else if (trn) begin 
		fsm_start_count <=  3'd4;
	end 
	else if (fsm_start_count == 3'd0) begin 
		fsm_start_count <= 3'd2;
	end 
	else begin 
		fsm_start_count <=  (fsm_start_count - 3'd1);
	end 

end 

assign fsm_start = (fsm_start_count == 3'd3) ? 1'b1 : 1'b0;
assign counter_en = (fsm_start);

always @(negedge spi_clk or posedge rst) begin 
	if (rst) begin 
		tx_count <=  5'd0;
	end 
	else if (counter_en) begin  
		tx_count <=  5'd24;
	end
	else if (tx_count == 5'd0) begin 
		tx_count <=  5'd24;
	end
	else begin   
		tx_count <=  tx_count - 5'd1;
	end  
end 

enum logic [2:0] {
  ST_IDLE      		= 3'd0,
  ST_TRN      		= 3'd1,
  ST_READ_DONE		= 3'd2
} curr_state, next_state;
 
always@(negedge spi_clk or posedge rst) begin
	if(rst == 1'b1)begin
		curr_state <=   ST_IDLE;
	end
	else if (SPI_EN == 1'b1) begin 
		curr_state <=   next_state;
	end
	else
	begin
		curr_state <=   ST_IDLE;
	end
end

always_comb begin 
	case(curr_state)
	ST_IDLE : 
	begin
		cs_z = 1'b1;
		busy = 1'b0;
		read_done_flag = 1'b0;
		tx_out = 1'b0;
		if (counter_en == 1'b1) begin 
			next_state = ST_TRN; 
		end 
		else begin next_state = ST_IDLE; end 
		
	end 
			
	
	ST_TRN : 
	begin   
		cs_z = 1'b0;
		read_done_flag = 1'b0;
		busy = 1'b1;
		tx_out = tx_buff[(tx_count - 5'd1)];  
		if (tx_count == 5'd1) begin next_state = ST_READ_DONE; end
		else begin next_state = ST_TRN; end 
				
	end

	ST_READ_DONE : 
	begin   
		cs_z = 1'b0;
		tx_out = 1'b0;
		busy = 1'b1;
		if ((tx_count == 5'd0) && (read_mode_en == 1'b1)) begin 
			read_done_flag = 1'b1; 
			next_state = ST_READ_DONE;
		end
		else begin 
			read_done_flag = 1'b0;
			next_state = ST_IDLE;
		end 
	end

        default	:
        begin
             cs_z = 1'b1;
		     busy = 1'b0;
		     read_done_flag = 1'b0;
		     tx_out = 1'b0;
             next_state = ST_IDLE;
        end 
		
	endcase

end

assign st_trn = (curr_state == ST_TRN) ? 1'b1 : 1'b0;
assign spi_clk_out = (st_trn) ? spi_clk : 1'b0;

// RX Shift reg

always @(posedge SPI_SCLK or posedge rst) begin 
	if(rst) begin 
		temp <=  24'h000000;
	end 
	else if (read_mode_en) begin 
		temp <= {temp[22:0],SPI_MISO};
	end 
	else begin 
		temp <=  24'h000000;
	end 
end 

assign rx_buff = temp;
		
// Read mode en logic 

always @(posedge rx_trn_pulse or negedge read_done_flag or posedge rst) begin 
	if (rx_trn_pulse) begin 
		read_mode_en <=  1'b1;
	end 
	else if (rst) begin 
	   read_mode_en  <= 1'b0;
	end 
	else if (!read_done_flag) begin 
		read_mode_en <=  1'b0;
	end 
	else begin 
		read_mode_en <=  1'b0;
	end 
end 

// read done logic 
always @(posedge read_done_flag or posedge counter_en or posedge rst) begin 
	if (rst) begin read_done <= 1'b0; end 
	else if (counter_en) begin 
		read_done <= 1'b0;
	end 
	else if (read_done_flag) begin 
		read_done <= 1'b1;
	end 
	else begin 
		read_done <= 1'b0;
	end 
end 

assign read_data = rx_buff;
assign SPI_MOSI = (!cs_z) ? (tx_out) : 1'b1; // Hi Z
assign SPI_SCLK = spi_clk_out;
assign SPI_CS_Z = cs_z;
assign SPI_BUSY = busy;
assign SPI_READ_DONE = read_done;

endmodule

